import abc
from datetime import datetime
from typing import ClassVar, Dict, List, Optional, Type, Callable

import numpy as np
from loguru import logger

from lcfm_shapefiles.labels import LCM10LabelsColors


class Product(abc.ABC):
    """Abstract base class for all products."""

    @property
    @abc.abstractmethod
    def product_name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def resolution(self) -> int:
        """Get the resolution of the product in meters."""
        pass
    
    @property
    @abc.abstractmethod
    def tile_size(self) -> int:
        """Get the tile size in degrees."""
        pass
        
    @property
    def layer_name(self):
        """Get the name of the layer."""
        return self._layer_name

    @property
    def bands(self):
        """Get the bands of the layer."""
        return self._bands

    @property
    def scale(self):
        """Get the scale of the layer."""
        return self._scale

    @property
    def offset(self):
        """Get the offset of the layer."""
        return self._offset
    
    @property
    def overview_method(self):
        """Get the overview method of the layer."""
        return "MODE" if self._layer_name == "MAP" else "BILINEAR"
    
    @property
    def product_type(self):
        return "Categorical" if self._layer_name == "MAP" else "Continuous"

    def __init__(self, version: str, layer_name: str, bands: List[str], scale: float|None = None, offset: float|None = None, year: int = 2020):
        """Initialize Product with a version and optional year.
        
        Args:
            version: Version string in the format 'vXXX' where X is a digit
            year: The year for the product, defaults to 2020
            layer_name: The name of the layer
            bands: The bands of the layer
            scale: The scale of the layer
            offset: The offset of the layer
        
        Raises:
            ValueError: If version doesn't follow the required format
        """
        # Assert that version follows the vxxx pattern
        if not version.startswith("v") or not version[1:].isdigit() or (len(version) != 4):
            logger.warning("Version must start with 'v' followed by 3 digits.")
        
        # Set the version and year
        self.version = version
        self.year = year
        
        # Set the layer properties
        self._layer_name = layer_name
        self._bands = bands
        self._scale = scale
        self._offset = offset

    def get_metadata(self, tile_id: str):
        """Generate metadata for a product.
        
        Args:
            tile_id: The tile identifier
            
        Returns:
            A dictionary containing the metadata
        """
        generation_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")

        metadata = dict(
            copyright=(
                f"LCFM project {self.year} / Contains modified "
                f"Copernicus Sentinel data ({self.year}) processed by "
                "LCFM consortium"
            ),
            creation_time=generation_time,
            description=f"Land Cover Map - {self.resolution}m",
            doi="undefined",
            license="CC-BY 4.0 - https://creativecommons.org/licenses/by/4.0/",
            platform="Sentinel-1, Sentinel-2",
            producer="VITO NV",
            product_crs="EPSG:4326",
            product_grid=f"{self.tile_size}x{self.tile_size} degree tiling grid",
            product_name=self.product_name,
            product_layer=self.layer_name,
            product_tile=tile_id,
            reference="https://land.copernicus.eu",
            sensor="SAR, MSI",
            time_start=f"{self.year}-01-01T00:00:00Z",
            time_end=f"{self.year}-12-31T23:59:59Z",
            version=self.version,
        )
        return metadata


class LCM10Product(Product):
    """Base class representing the Land Cover Map 10m product."""
    # Using ClassVar to indicate these are class variables, not instance variables
    nodata: ClassVar[int] = LCM10LabelsColors.NO_DATA.id
    data_type: ClassVar[type] = np.uint8
    product_name: ClassVar[str] = "LCM-10"
    
    # Registry to store layer subclasses specific to LCM10
    _registry: ClassVar[Dict[str, Type['LCM10Product']]] = {}
    
    @classmethod
    def register(cls, band_prefix: str):
        """Register a product class for a specific band prefix.
        
        Args:
            band_prefix: The prefix used by the bands of this layer
            
        Returns:
            A decorator function that registers the class
        """
        def decorator(subclass):
            cls._registry[band_prefix.upper()] = subclass
            return subclass
        return decorator
    
    @classmethod
    def create_layer(cls, version: str, band: str, year: int = 2020) -> 'LCM10Product':
        """Create a layer instance based on a band name.
        
        Args:
            version: The version string
            band: The band name (will match with band prefix)
            year: The year for the product
            
        Returns:
            An instance of the appropriate LCM10 product subclass
            
        Raises:
            ValueError: If no matching layer is found for the band
        """
        band = band.upper()
        
        # First try for an exact match
        if band in cls._registry:
            return cls._registry[band](version=version, year=year)
        
        # Sort prefixes by length (descending) to match the most specific prefix first
        sorted_prefixes = sorted(cls._registry.keys(), key=len, reverse=True)
        
        # Then try for prefix matches
        for prefix in sorted_prefixes:
            if band.startswith(prefix):
                return cls._registry[prefix](version=version, year=year)
                
        registered_prefixes = ", ".join(cls._registry.keys())
        raise ValueError(f"Unknown LCM10 band: {band}. Available band prefixes: {registered_prefixes}")
    
    @property
    def resolution(self) -> int:
        """Get the resolution of the product in meters."""
        return 10
    
    @property
    def tile_size(self) -> int:
        """Get the tile size in degrees."""
        return 3


@LCM10Product.register("MAP")
class LCM10MapProduct(LCM10Product):
    """Product class for LCM10 MAP product."""
    
    def __init__(self, version: str, year: int = 2020):
        super().__init__(
            version=version, 
            year=year,
            layer_name="MAP",
            bands=["MAP"]
        )

    def get_metadata(self, tile_id: str):
        """Generate metadata for a product.
        
        Args:
            tile_id: The tile identifier
            
        Returns:
            A dictionary containing the metadata
        """
        metadata = super().get_metadata(tile_id)
        
        # Generate legend dynamically based on LCM10LabelsColors
        legend_items = []
        for label in LCM10LabelsColors:
            if label.id != LCM10LabelsColors.NO_DATA.id:
                legend_items.append(f"{label.id} - {label.name}")
        metadata["legend"] = "\n".join(legend_items)
        return metadata

@LCM10Product.register("MAP-CONTINUOUS")
@LCM10Product.register("MAPRGB")
class LCM10MapContinuousProduct(LCM10Product):
    """Product class for LCM10 MAP-CONTINUOUS product."""
    
    def __init__(self, version: str, year: int = 2020):
        super().__init__(
            version=version, 
            year=year,
            layer_name="MAP-CONTINUOUS",
            bands=["MAP-CONTINUOUS"]
        )


@LCM10Product.register("PROB")
class LCM10ProbabilityProduct(LCM10Product):
    """Product class for LCM10 PROBABILITY product."""
    
    def __init__(self, version: str, year: int = 2020):
        # Generate probability bands based on actual label IDs from LCM10LabelsColors
        valid_labels = [label for label in LCM10LabelsColors if label.id != LCM10LabelsColors.NO_DATA.id and label.id != LCM10LabelsColors.UNCLASSIFIABLE.id]
        bands = [f"PROB{label.id}" for label in valid_labels]
        
        super().__init__(
            version=version, 
            year=year,
            layer_name="PROBABILITY",
            bands=bands,
            scale=0.004
        )


class LCM100Product(Product):
    """Base class representing the Land Cover Map 100m product."""
    # Using ClassVar to indicate these are class variables, not instance variables
    nodata: ClassVar[int] = 255
    data_type: ClassVar[type] = np.uint8
    product_name: ClassVar[str] = "LCM-100"
    
    # Registry to store layer subclasses specific to LCM100
    _registry: ClassVar[Dict[str, Type['LCM100Product']]] = {}
    
    @classmethod
    def register(cls, band_prefix: str):
        """Register a product class for a specific band prefix.
        
        Args:
            band_prefix: The prefix used by the bands of this layer
            
        Returns:
            A decorator function that registers the class
        """
        def decorator(subclass):
            cls._registry[band_prefix.upper()] = subclass
            return subclass
        return decorator
    
    @classmethod
    def create_layer(cls, version: str, band: str, year: int = 2020) -> 'LCM100Product':
        """Create a layer instance based on a band name.
        
        Args:
            version: The version string
            band: The band name (will match with band prefix)
            year: The year for the product
            
        Returns:
            An instance of the appropriate LCM100 product subclass
            
        Raises:
            ValueError: If no matching layer is found for the band
        """
        band = band.upper()
        
        # First try for an exact match
        if band in cls._registry:
            return cls._registry[band](version=version, year=year)
        
        # Sort prefixes by length (descending) to match the most specific prefix first
        sorted_prefixes = sorted(cls._registry.keys(), key=len, reverse=True)
        
        # Then try for prefix matches
        for prefix in sorted_prefixes:
            if band.startswith(prefix):
                return cls._registry[prefix](version=version, year=year)
                
        registered_prefixes = ", ".join(cls._registry.keys())
        raise ValueError(f"Unknown LCM100 band: {band}. Available band prefixes: {registered_prefixes}")
    
    @property
    def resolution(self) -> int:
        """Get the resolution of the product in meters."""
        return 100
    
    @property
    def tile_size(self) -> int:
        """Get the tile size in degrees."""
        return 30


@LCM100Product.register("MAP")
class LCM100MapProduct(LCM100Product):
    """Product class for LCM100 MAP product."""
    
    def __init__(self, version: str, year: int = 2020):
        super().__init__(
            version=version, 
            year=year,
            layer_name="MAP",
            bands=["MAP"]
        )

    def get_metadata(self, tile_id: str):
        """Generate metadata for a product.
        
        Args:
            tile_id: The tile identifier
            
        Returns:
            A dictionary containing the metadata
        """
        metadata = super().get_metadata(tile_id)
        # Generate legend dynamically based on LCM10LabelsColors
        legend_items = []
        legend_items.append("11 - Closed tree cover")
        legend_items.append("12 - Open tree cover")
        for label in LCM10LabelsColors:
            if label.id != LCM10LabelsColors.NO_DATA.id and label.id != LCM10LabelsColors.TREE.id:
                legend_items.append(f"{label.id} - {label.name}")
        metadata["legend"] = "\n".join(legend_items)
        return metadata


@LCM100Product.register("CFRAC")
class LCM100CFracProduct(LCM100Product):
    """Product class for LCM100 CFRAC product."""
    
    def __init__(self, version: str, year: int = 2020):
        CFRAC_NAMES = [
            "TREECOVER",
            "SHRUBLAND",
            "GRASSLAND",
            "CROPLAND",
            "BUILTUP",
            "BARE",
            "SNOW",
            "WATER",
            "WETLAND",
            "MANGROVES",
            "LICHENS",
        ]
        bands = [f"CFRAC_{name}" for name in CFRAC_NAMES]
        super().__init__(
            version=version, 
            year=year,
            layer_name="CFRAC",
            bands=bands,
            scale=0.004
        )


@LCM100Product.register("CHANGE")
class LCM100ChangeProduct(LCM100Product):
    """Product class for LCM100 CHANGE product."""
    
    def __init__(self, version: str, year: int = 2020):
        super().__init__(
            version=version, 
            year=year,
            layer_name="CHANGE",
            bands=["CHANGE"]
        )
